@@ -1,74 +1,292 @@
+from __future__ import annotations
+
 from abc import ABC, abstractmethod
-from typing import Collection, Optional, Set, Tuple
+from typing import Callable, Iterator, List, Optional, overload, Set, Tuple
 
-from .types import Element, Morphism, Object
+from .helper import IOHelper
+from .types import ConcreteRepr, Element, Morphism, Object
+
+__all__ = [
+    "InvalidFormat",
+    "InvalidValue",
+    "EnumerableSetsOperations",
+    "EnumerableSet",
+    "FiniteSetUnion",
+    "Setoid",
+    "SetoidOperations",
+    "SetProduct",
+    "SetDisjointUnion",
+    "MakeSetUnion",
+    "MakeSetDisjointUnion",
+    "MakeSetIntersection",
+    "MakeSetProduct",
+    "MonotoneMap",
+    "Mapping",
+    "Monoid",
+    "MonoidalCategory",
+    "MonoidalCategory",
+    "FiniteMap",
+    "FiniteDP",
+    "FiniteSet",
+    "FiniteSetProperties",
+    "FiniteSetProduct",
+    "FiniteProfunctor",
+    "FiniteSetDisjointUnion",
+    "FiniteFunctor",
+    "FiniteSemigroupConstruct",
+    "FiniteSemigroup",
+    "FinitePoset",
+    "FiniteAdjunction",
+    "FiniteMonoidalPoset",
+    "FiniteMonoidalCategory",
+    "FiniteCategory",
+    "FiniteEnrichedCategory",
+    "FiniteRelation",
+    "FiniteSemiCategory",
+    "FiniteMonoid",
+    "FiniteCategoryOperations",
+    "FiniteDPOperations",
+    "FiniteAdjunctionsOperations",
+    "FiniteProfunctorOperations",
+    "FiniteGroup",
+    "FiniteLattice",
+    "FiniteRelationProperties",
+    "FiniteRelationOperations",
+    "FiniteEndorelationOperations",
+    "FiniteMonotoneMapProperties",
+    "FiniteEndorelationProperties",
+    "FinitePosetOperations",
+    "FinitePosetConstructors",
+    "FinitePosetSubsetProperties",
+    "FinitePosetProperties",
+    "FiniteMapOperations",
+    "FinitePosetSubsetOperations",
+    "FiniteNaturalTransformation",
+    "Semigroup",
+    "Lattice", "JoinSemilattice", "Adjunction",
+    'MeetSemilattice', "MonoidalPosetOperations", "SemiCategory",
+    "SemiBiCategory", "Poset", "Category", "CategoryOperations", 'DP', 'DPCategory', 'DPI',
+    "DPConstructors", "InvalidFormat",
+]
 
 
-class FiniteSet(ABC):
+class Setoid(ABC):
+    """
+    A setoid is something to which elements may belong,
+    has a way of distinguishing elements,
+    and is able to (de)serialize its elements.
+    """
+
+    @abstractmethod
+    def contains(self, x: Element) -> bool:
+        """ Returns true if the element is in the set. """
+
+    def equal(self, x: Element, y: Element) -> bool:
+        """ Returns True if the two elements are to be considered equal. """
+        return x == y  # default is to use the Python equality
+
+    def apart(self, x: Element, y: Element) -> bool:
+        return not self.equal(x, y)
+
+    @abstractmethod
+    def save(self, h: IOHelper, x: Element) -> ConcreteRepr:
+        ...
+
+    @abstractmethod
+    def load(self, h: IOHelper, o: ConcreteRepr) -> Element:
+        ...
+
+
+class Mapping(ABC):
+    @abstractmethod
+    def source(self) -> Setoid:
+        ...
+
+    @abstractmethod
+    def target(self) -> Setoid:
+        ...
+
+    @abstractmethod
+    def __call__(self, a: Element) -> Element:
+        ...
+
+
+class EnumerableSet(Setoid, ABC):
+    @abstractmethod
+    def elements(self) -> Iterator[Element]:
+        """ Note: possibly non-terminating. """
+
+
+class FiniteSet(EnumerableSet, ABC):
     """ A finite set has a finite size. """
 
     @abstractmethod
     def size(self) -> int:
         """ Return the size of the finite set. """
 
+
+class FiniteMap(Mapping, ABC):
     @abstractmethod
-    def belongs(self, x: Element) -> bool:
+    def source(self) -> FiniteSet:
         ...
 
     @abstractmethod
-    def equal(self, x: Element, y: Element) -> bool:
-        """ Returns True if the two elements are to be considered equal. """
-
-    @abstractmethod
-    def elements(self) -> Collection[Element]:
+    def target(self) -> FiniteSet:
         ...
 
 
 class InvalidFormat(Exception):
-    pass
+    """ Raise this if the input data to parse is invalid. """
 
 
-class FiniteSetRepresentation(ABC):
-
-    @abstractmethod
-    def load(self, yaml_data: str) -> FiniteSet:
-        """ Load a finite set from given YAML data.
-            Throw InvalidFormat if the format is incorrect.
-        """
-
-    @abstractmethod
-    def save(self, f: FiniteSet) -> str:
-        """ Save to a string. """
+class InvalidValue(ValueError):
+    """ Raise this if the input does not make sense. """
 
 
 class FiniteSetProperties(ABC):
-
     @abstractmethod
     def is_subset(self, a: FiniteSet, b: FiniteSet) -> bool:
-        ...
+        """ True if `a` is a subset of `b`. """
+
+    def is_equal(self, a: FiniteSet, b: FiniteSet) -> bool:
+        return self.is_subset(a, b) and self.is_subset(b, a)
+
+    def is_strict_subset(self, a: FiniteSet, b: FiniteSet) -> bool:
+        return self.is_subset(a, b) and not self.is_subset(b, a)
 
 
-class FiniteSetOperations(ABC):
+class SetProduct(Setoid, ABC):
+    """ A set product is a setoid that can be factorized. """
 
     @abstractmethod
-    def union(self, s1: FiniteSet, s2: FiniteSet) -> FiniteSet:
+    def components(self) -> List[Setoid]:
+        """ Returns the components of the product"""
+
+    @abstractmethod
+    def pack(self, *args: Element) -> Element:
+        """ Packs an element of each setoid into an element of the mapping"""
+
+    @abstractmethod
+    def projections(self) -> List[Mapping]:
+        """ Returns the projection mappings. """
+
+
+class FiniteSetProduct(FiniteSet, SetProduct, ABC):
+    """ Specialization of SetProduct where we deal with FiniteSets"""
+
+    @abstractmethod
+    def components(self) -> List[FiniteSet]:
+        """ Returns the components """
+
+    @abstractmethod
+    def projections(self) -> List[FiniteMap]:
+        """ Returns the projection mappings. """
+
+
+class SetUnion(Setoid, ABC):
+    """ A set product is a setoid that can be factorized. """
+
+    @abstractmethod
+    def components(self) -> List[Setoid]:
+        """ Returns the components of the union"""
+
+
+class EnumerableSetUnion(EnumerableSet, SetUnion, ABC):
+    """ Specialization of SetUnion where we deal with FiniteSets"""
+
+    @abstractmethod
+    def components(self) -> List[EnumerableSet]:
+        """ Returns the components of the union """
+
+
+class FiniteSetUnion(FiniteSet, EnumerableSetUnion, ABC):
+    """ Specialization of SetUnion where we deal with FiniteSets"""
+
+    @abstractmethod
+    def components(self) -> List[FiniteSet]:
+        """ Returns the components of the union """
+
+
+class SetDisjointUnion(Setoid, ABC):
+    @abstractmethod
+    def components(self) -> List[Setoid]:
+        """ Returns the components of the union """
+
+    @abstractmethod
+    def injections(self) -> List[Mapping]:
+        """ Returns the projection mappings. """
+
+
+class FiniteSetDisjointUnion(FiniteSet, SetDisjointUnion, ABC):
+    """ Specialization of SetProduct where we deal with FiniteSets"""
+
+    @abstractmethod
+    def components(self) -> List[FiniteSet]:
         ...
 
     @abstractmethod
-    def intersection(self, s1: FiniteSet, s2: FiniteSet) -> FiniteSet:
+    def injections(self) -> List[FiniteMap]:
+        ...
+
+
+class MakeSetProduct(ABC):
+    @overload
+    def product(self, components: List[Setoid]) -> SetProduct:
         ...
 
     @abstractmethod
-    def product(self, s1: FiniteSet, s2: FiniteSet) -> FiniteSet:
+    def product(self, components: List[FiniteSet]) -> FiniteSetProduct:
+        ...
+
+
+class MakeSetIntersection(ABC):
+    @abstractmethod
+    def intersection(self, components: List[FiniteSet]) -> FiniteSet:
+        ...
+
+
+class MakeSetUnion(ABC):
+    @overload
+    def union(self, components: List[FiniteSet]) -> FiniteSetUnion:
+        ...
+
+    @overload
+    def union(self, components: List[EnumerableSet]) -> EnumerableSetUnion:
         ...
 
     @abstractmethod
-    def disjoint_union(self, s1: FiniteSet, s2: FiniteSet) -> FiniteSet:
+    def union(self, components: List[Setoid]) -> SetUnion:
         ...
 
 
-class FiniteRelation(ABC):
+# added
+class MakeSetDisjointUnion(ABC):
+    @overload
+    def compute_disjoint_union(self, components: List[Setoid]) -> SetDisjointUnion:
+        ...
 
+    @abstractmethod
+    def compute_disjoint_union(
+        self, components: List[FiniteSet]
+    ) -> FiniteSetDisjointUnion:
+        ...
+
+
+class Relation(ABC):
+    @abstractmethod
+    def source(self) -> Setoid:
+        """ Returns a setoid """
+
+    @abstractmethod
+    def target(self) -> Setoid:
+        """ Returns a setoid """
+
+    @abstractmethod
+    def holds(self, a: Element, b: Element) -> bool:
+        """ Returns true if the two elements are related """
+
+
+class FiniteRelation(Relation, ABC):
     @abstractmethod
     def source(self) -> FiniteSet:
         """ Returns a finite set"""
@@ -77,24 +295,8 @@ class FiniteRelation(ABC):
     def target(self) -> FiniteSet:
         """ Returns a finite set"""
 
-    @abstractmethod
-    def holds(self, a: Element, b: Element) -> bool:
-        """ Returns true if the two elements are related """
-
-
-class FiniteRelationRepresentation(ABC):
-
-    @abstractmethod
-    def load(self, yaml_data: str) -> FiniteRelation:
-        """ Load a finite set from given YAML data"""
-
-    @abstractmethod
-    def save(self, f: FiniteRelation) -> str:
-        """ Load a finite set from given YAML data"""
-
 
 class FiniteRelationProperties(ABC):
-
     @abstractmethod
     def is_surjective(self, fr: FiniteRelation) -> bool:
         """ Return True if the relation is surjective. """
@@ -113,14 +315,12 @@ class FiniteRelationProperties(ABC):
 
 
 class FiniteRelationOperations(ABC):
-
     @abstractmethod
     def transpose(self, fr: FiniteRelation) -> FiniteRelation:
         """ Create the transposed of a relation """
 
 
 class FiniteEndorelationProperties(ABC):
-
     @abstractmethod
     def is_reflexive(self, fr: FiniteRelation) -> bool:
         """ Return True if the relation is reflexive. """
@@ -147,40 +347,12 @@ class FiniteEndorelationProperties(ABC):
 
 
 class FiniteEndorelationOperations(ABC):
-
     @abstractmethod
     def transitive_closure(self, fr: FiniteRelation) -> FiniteRelation:
         """ Returns transitive closure """
 
 
-class FiniteMap(ABC):
-
-    @abstractmethod
-    def source(self) -> FiniteSet:
-        """ Returns a finite set"""
-
-    @abstractmethod
-    def target(self) -> FiniteSet:
-        """ Returns a finite set"""
-
-    @abstractmethod
-    def exec(self, a: Element) -> Element:
-        ...
-
-
-class FiniteMapRepresentation(ABC):
-
-    @abstractmethod
-    def load(self, s: str) -> FiniteMap:
-        """ Load the data  """
-
-    @abstractmethod
-    def save(self, m: FiniteMap) -> str:
-        """ Load the data  """
-
-
 class FiniteMapOperations(ABC):
-
     @abstractmethod
     def compose(self, f: FiniteMap, g: FiniteMap) -> FiniteMap:
         """ compose two functions"""
@@ -190,10 +362,9 @@ class FiniteMapOperations(ABC):
         """ Load the data  """
 
 
-class FiniteSemigroup(ABC):
-
+class Semigroup(ABC):
     @abstractmethod
-    def carrier(self) -> FiniteSet:
+    def carrier(self) -> Setoid:
         ...
 
     @abstractmethod
@@ -201,69 +372,60 @@ class FiniteSemigroup(ABC):
         ...
 
 
-class FiniteSemigroupRepresentation(ABC):
-
+class FiniteSemigroup(Semigroup, ABC):
     @abstractmethod
-    def load(self, s: str) -> FiniteSemigroup:
-        """ Load the data  """
-
-    @abstractmethod
-    def save(self, m: FiniteSemigroup) -> str:
-        """ Save the data  """
+    def carrier(self) -> FiniteSet:
+        ...
 
 
 class FiniteSemigroupConstruct(ABC):
-
     @abstractmethod
     def free(self, fs: FiniteSet) -> FiniteSemigroup:
         """ Construct the free semigroup on a set. """
 
 
-class FiniteMonoid(FiniteSemigroup, ABC):
-
+class Monoid(Semigroup, ABC):
     @abstractmethod
     def identity(self) -> Element:
         ...
 
 
-class FiniteMonoidRepresentation(ABC):
+class Group(Monoid, ABC):
+    @abstractmethod
+    def inverse(self, e: Element) -> Element:
+        """ Returns the inverse of an element"""
+
+
+class FiniteMonoid(Monoid, FiniteSemigroup, ABC):
+    """"""
+
+
+class FiniteGroup(Group, FiniteMonoid, ABC):
+    ...
+
+
+# TODO: equational theories
+
+
+class Poset(ABC):
+    @abstractmethod
+    def carrier(self) -> Setoid:
+        ...
 
     @abstractmethod
-    def load(self, s: str) -> FiniteMonoid:
-        """ Load the data  """
-
-    @abstractmethod
-    def save(self, m: FiniteMonoid) -> str:
-        """ Save the data  """
+    def leq(self, a: Element, b: Element) -> bool:
+        ...
 
 
-# TODO: equational theotires
-
-class FinitePoset(ABC):
+class FinitePoset(Poset, ABC):
     """ Implementation of finite posets. """
 
     @abstractmethod
     def carrier(self) -> FiniteSet:
         ...
 
-    @abstractmethod
-    def leq(self, x: Element, y: Element) -> bool:
-        """ Implements $\posleq$ """
-
-
-class FinitePosetRepresentation(ABC):
-
-    @abstractmethod
-    def load(self, s: str) -> FinitePoset:
-        """ Load the data  """
-
-    @abstractmethod
-    def save(self, m: FinitePoset) -> str:
-        """ Save the data  """
-
 
 class FinitePosetProperties(ABC):
-
     @abstractmethod
     def width(self, fp: FinitePoset) -> int:
         """ Return the width of the poset. """
@@ -274,7 +436,6 @@ class FinitePosetProperties(ABC):
 
 
 class FinitePosetConstructors(ABC):
-
     @abstractmethod
     def discrete(self, s: FiniteSet) -> FinitePoset:
         """ Creates the discrete poset from any set. """
@@ -305,7 +466,6 @@ class FinitePosetConstructors(ABC):
 
 
 class FinitePosetSubsetProperties(ABC):
-
     @abstractmethod
     def is_chain(self, fp: FinitePoset, s: FiniteSet) -> bool:
         """ Computes if the subset is a chain. """
@@ -316,7 +476,6 @@ class FinitePosetSubsetProperties(ABC):
 
 
 class FinitePosetSubsetOperations(ABC):
-
     @abstractmethod
     def upperclosure(self, fp: FinitePoset, s: Set[Element]) -> Set[Element]:
         """ Computes the upper closure of an element"""
@@ -351,7 +510,6 @@ class FinitePosetSubsetOperations(ABC):
 
 
 class FinitePosetOperations(ABC):
-
     @abstractmethod
     def opposite(self, s: str) -> FinitePoset:
         ...
@@ -365,16 +523,15 @@ class FinitePosetOperations(ABC):
         ...
 
 
-class FiniteMonotoneMap(ABC):
-    def source(self) -> FinitePoset:
+class MonotoneMap(Mapping, ABC):
+    def source_poset(self) -> Poset:
         ...
 
-    def target(self) -> FinitePoset:
+    def target_poset(self) -> Poset:
         ...
 
 
 class FiniteMonotoneMapProperties(ABC):
-
     @abstractmethod
     def is_monotone(self, p1: FinitePoset, p2: FinitePoset, m: FiniteMap) -> bool:
         """ Check if a map is monotone. """
@@ -404,7 +561,7 @@ class MonoidalPosetOperations(ABC):
         """ Check that the pair of poset and monoid make together a monoidal poset."""
 
 
-class FiniteMeetSemilattice:
+class MeetSemilattice(Poset, ABC):
     @abstractmethod
     def meet(self, x: Element, y: Element) -> Element:
         ...
@@ -414,7 +571,7 @@ class FiniteMeetSemilattice:
         ...
 
 
-class FiniteJoinSemilattice:
+class JoinSemilattice(Poset, ABC):
     @abstractmethod
     def join(self, x: Element, y: Element) -> Element:
         ...
@@ -424,12 +581,45 @@ class FiniteJoinSemilattice:
         ...
 
 
-class FiniteLattice(FiniteMeetSemilattice, FiniteJoinSemilattice, ABC):
+class Lattice(JoinSemilattice, MeetSemilattice, ABC):
+    pass
+
+
+class FiniteLattice(ABC):
+    @abstractmethod
+    def carrier(self) -> FiniteSet:
+        ...
+
+
+class SemiBiCategory(ABC):
+    @abstractmethod
+    def objects(self) -> Setoid:
+        ...
+
+    @abstractmethod
+    def homs(self, ob1: Object, ob2: Object) -> Setoid:
+        ...
+
+    @abstractmethod
+    def legs(self, m: Morphism) -> Tuple[Object, Object]:
+        """ Return source and target of the morphism """
+
+
+class SemiCategory(SemiBiCategory, ABC):
     ...
 
 
-class FiniteSemiCategory(ABC):
+class Category(SemiCategory, ABC):
+    @abstractmethod
+    def identity(self, ob: Object) -> Morphism:
+        """ Identity for the object """
 
+
+assert [1] + [1, 2] == [1, 1, 2]
+assert [1, 1, 2] + [] == [1, 1, 2]
+
+
+class FiniteSemiCategory(SemiCategory, ABC):
     @abstractmethod
     def objects(self) -> FiniteSet:
         ...
@@ -438,20 +628,30 @@ class FiniteSemiCategory(ABC):
     def homs(self, ob1: Object, ob2: Object) -> FiniteSet:
         ...
 
+
+class FiniteCategory(FiniteSemiCategory, Category, ABC):
+    ...
+
+
+class CategoryOperations:
     @abstractmethod
-    def legs(self, m: Morphism) -> Tuple[Object, Object]:
-        """ Return source and target of the morphism """
-
-
-class FiniteCategory(FiniteSemiCategory, ABC):
+    def product(self, c1: Category, c2: Category) -> Category:
+        """ Product of two categories. """
 
     @abstractmethod
-    def identity(self, ob: Object) -> Morphism:
-        """ Identity for the object """
+    def disjoint_union(self, c1: Category, c2: Category) -> FiniteCategory:
+        """ Disjoint union for the categories """
+
+    @abstractmethod
+    def arrow(self, c1: Category) -> Category:
+        """ Computes the arrow category """
+
+    @abstractmethod
+    def twisted_arrow(self, c1: Category) -> Category:
+        """ Computes the twisted arrow category """
 
 
 class FiniteCategoryOperations:
-
     @abstractmethod
     def product(self, c1: FiniteCategory, c2: FiniteCategory) -> FiniteCategory:
         """ Product of two categories. """
@@ -469,14 +669,13 @@ class FiniteCategoryOperations:
         """ Computes the twisted arrow category """
 
 
-class FiniteFunctor(ABC):
-
+class Functor(ABC):
     @abstractmethod
-    def source(self) -> FiniteCategory:
+    def source(self) -> Category:
         ...
 
     @abstractmethod
-    def target(self) -> FiniteCategory:
+    def target(self) -> Category:
         ...
 
     @abstractmethod
@@ -488,19 +687,17 @@ class FiniteFunctor(ABC):
         """ Effect on morphisms """
 
 
-class FiniteFunctorRepresentation(ABC):
+class FiniteFunctor(Functor, ABC):
+    @abstractmethod
+    def source(self) -> FiniteCategory:
+        ...
 
     @abstractmethod
-    def load(self, yaml_data: str) -> FiniteFunctor:
-        """ Load a functor from given YAML data"""
-
-    @abstractmethod
-    def save(self, f: FiniteFunctor) -> str:
+    def target(self) -> FiniteCategory:
         ...
 
 
-class FiniteMonoidalCategory(FiniteCategory, ABC):
-
+class MonoidalCategory(Category, ABC):
     @abstractmethod
     def monoidal_unit(self) -> Object:
         """ Return the product functor. """
@@ -510,8 +707,27 @@ class FiniteMonoidalCategory(FiniteCategory, ABC):
         """ Return the product functor. """
 
 
-class FiniteNaturalTransformation(ABC):
+class FiniteMonoidalCategory(MonoidalCategory, FiniteCategory, ABC):
+    ...
 
+
+class NaturalTransformation(ABC):
+    @abstractmethod
+    def cat1(self) -> Category:
+        ...
+
+    @abstractmethod
+    def cat2(self) -> Category:
+        ...
+
+    @abstractmethod
+    def component(self, ob: Object) -> Morphism:
+        """Returns the component for a particular object in the first category.
+        This is a morphism in the second category.
+        """
+
+
+class FiniteNaturalTransformation(NaturalTransformation, ABC):
     @abstractmethod
     def cat1(self) -> FiniteCategory:
         ...
@@ -520,26 +736,26 @@ class FiniteNaturalTransformation(ABC):
     def cat2(self) -> FiniteCategory:
         ...
 
+
+class Adjunction(ABC):
     @abstractmethod
-    def component(self, ob: Object) -> Morphism:
-        """ Returns the component for a particular object in the first category.
-            This is a morphism in the second category.
-        """
-
-
-class FiniteNaturalTransformationRepresentation(ABC):
-
-    @abstractmethod
-    def load(self, yaml_data: str) -> FiniteNaturalTransformation:
-        """ Load a natural transformation from given YAML data"""
-
-    @abstractmethod
-    def save(self, f: FiniteNaturalTransformation) -> str:
+    def source(self) -> Category:
         ...
 
+    @abstractmethod
+    def target(self) -> Category:
+        ...
 
-class FiniteAdjunction(ABC):
+    @abstractmethod
+    def left(self) -> Functor:
+        pass
 
+    @abstractmethod
+    def right(self) -> Functor:
+        pass
+
+
+class FiniteAdjunction(Adjunction, ABC):
     @abstractmethod
     def source(self) -> FiniteCategory:
         ...
@@ -557,27 +773,15 @@ class FiniteAdjunction(ABC):
         pass
 
 
-class FiniteAdjunctionRepresentation(ABC):
-
-    @abstractmethod
-    def load(self, yaml_data: str) -> FiniteAdjunction:
-        ...
-
-    @abstractmethod
-    def save(self, f: FiniteAdjunction) -> str:
-        ...
-
-
 class FiniteAdjunctionsOperations(ABC):
-
     @abstractmethod
-    def is_adjunction(self, left: FiniteFunctor,
-                      right: FiniteFunctor) -> bool:
+    def is_adjunction(self, left: FiniteFunctor, right: FiniteFunctor) -> bool:
         """ check the pair is an adjunction """
 
     @abstractmethod
-    def compose(self, adj1: FiniteAdjunction,
-                adj2: FiniteAdjunction) -> FiniteAdjunction:
+    def compose(
+        self, adj1: FiniteAdjunction, adj2: FiniteAdjunction
+    ) -> FiniteAdjunction:
         """ compose two compatible adjunctions"""
 
     @abstractmethod
@@ -585,34 +789,51 @@ class FiniteAdjunctionsOperations(ABC):
         ...
 
 
+class DPI(ABC):
+    @abstractmethod
+    def functionality(self) -> Poset:
+        ...
+
+    @abstractmethod
+    def implementations(self) -> Setoid:
+        ...
+
+    @abstractmethod
+    def costs(self) -> Poset:
+        ...
+
+    @abstractmethod
+    def requires(self) -> Mapping:
+        ...
+
+    @abstractmethod
+    def provides(self) -> Mapping:
+        ...
+
+
+class DPCategory(Category, ABC):
+    pass
+
+
+class DP(ABC):
+    pass
+
+
 class FiniteDP(ABC):
     pass
 
 
-class FiniteDPRepresentation(ABC):
-
+class DPConstructors(ABC):
     @abstractmethod
-    def load(self, yaml_data: str) -> FiniteDP:
-        ...
-
-    @abstractmethod
-    def save(self, f: FiniteDP) -> str:
-        ...
-
-
-class FiniteDPConstructors(ABC):
-
-    @abstractmethod
-    def companion(self, f: FiniteMonotoneMap) -> FiniteDP:
+    def companion(self, f: MonotoneMap) -> DP:
         pass
 
     @abstractmethod
-    def conjoint(self, f: FiniteMonotoneMap) -> FiniteDP:
+    def conjoint(self, f: MonotoneMap) -> DP:
         pass
 
 
 class FiniteDPOperations(ABC):
-
     @abstractmethod
     def series(self, dp1: FiniteDP, dp2: FiniteDP) -> FiniteDP:
         pass
@@ -630,6 +851,17 @@ class FiniteDPOperations(ABC):
         ...
 
 
+class Profunctor(ABC):
+    def source(self) -> Category:
+        ...
+
+    def target(self) -> Category:
+        ...
+
+    def functor(self) -> Functor:
+        ...
+
+
 class FiniteProfunctor(ABC):
     def cat1(self) -> FiniteCategory:
         ...
@@ -641,19 +873,7 @@ class FiniteProfunctor(ABC):
         ...
 
 
-class FiniteProfunctorRepresentation(ABC):
-
-    @abstractmethod
-    def load(self, yaml_data: str) -> FiniteProfunctor:
-        """ Load a natural transformation from given YAML data"""
-
-    @abstractmethod
-    def save(self, f: FiniteProfunctor) -> str:
-        ...
-
-
 class FiniteProfunctorOperations(ABC):
-
     @abstractmethod
     def series(self, p1: FiniteProfunctor, p2: FiniteProfunctor) -> FiniteProfunctor:
         ...
@@ -662,3 +882,28 @@ class FiniteProfunctorOperations(ABC):
 class FiniteEnrichedCategory(FiniteCategory, ABC):
     def enrichment(self) -> FiniteMonoidalCategory:
         ...
+
+
+class SetoidOperations(ABC):
+    @classmethod
+    @abstractmethod
+    def union_setoids(cls, a: Setoid, b: Setoid) -> Setoid:
+        """ Creates the union of two Setoids. """
+
+    @classmethod
+    @abstractmethod
+    def intersection_setoids(cls, a: Setoid, b: Setoid) -> Setoid:
+        """ Creates the intersection of two Setoids. """
+
+
+class EnumerableSetsOperations(ABC):
+    @classmethod
+    @abstractmethod
+    def make_set_sequence(cls, f: Callable[[int], object]):
+        """Creates an EnumerableSet from a function that gives the
+        i-th element."""
+
+    @classmethod
+    @abstractmethod
+    def union_esets(cls, a: EnumerableSet, b: EnumerableSet) -> EnumerableSet:
+        """ Creates the union of two EnumerableSet. """
